import numpy as np

from skbio.stats.composition import multiplicative_replacement, closure, clr

from sklearn.base import TransformerMixin, BaseEstimator, clone
from sklearn.feature_selection import VarianceThreshold
from sklearn.utils import resample


class CLRClosureTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, do_clr=False, delta=None):

        self.do_clr = do_clr
        self.delta = delta

    def fit_transform(self, X, y=None, **kwargs):

        if self.do_clr:
            return clr(multiplicative_replacement(closure(X), delta=self.delta)).astype(np.float32)

        else:
            return closure(X).astype(np.float32)

    def transform(self, X, y=None, **kwargs):

        if self.do_clr:
            return clr(multiplicative_replacement(closure(X), delta=self.delta)).astype(np.float32)

        else:
            return closure(X).astype(np.float32)


class ResampleRandomizeTransform(BaseEstimator, TransformerMixin):
    def __init__(self, resampler, transformer, exclude_cols):

        self.resampler = resampler
        self.transformer = transformer
        self.exclude_cols = exclude_cols

    def fit_resample(self, X, y):
        
        # Resample data if a resampler is supplied
        if isinstance(self.resampler, type(None)):
            X_re = np.copy(X)

        else:
            self.resampler = clone(self.resampler)

            X_re, _ = self.resampler.fit_resample(X, y)

        y_re = np.asarray([0 for _ in range(X_re.shape[0])])

        # Create Randomized Data
        X_perm = np.copy(X_re, "C")
        for col in range(X_perm.shape[1]):
            X_perm[:, col] = np.random.choice(
                X_perm[:, col], replace=False, size=X_perm.shape[0]
            )

        y_ra = np.asarray([1 for _ in range(X_perm.shape[0])])

        # Combine the real and permuted data
        X_comb = np.vstack((X_re, X_perm))

        y_comb = np.hstack((y_re, y_ra))

        # Prepare dataset if columns are excluded from transformation
        if self.exclude_cols[0]:
            # Split data into columns excluded from transformation and included for transformation
            excl_range = np.asarray(self.exclude_cols[1])

            X_transform = np.delete(X_comb, excl_range, axis=1)
            X_no_transform = X_comb[:, excl_range]

            # Remove zero-variance features
            self.zero_var_transformer = VarianceThreshold().fit(X_transform)
            self.zero_var_no_transformer = VarianceThreshold().fit(X_no_transform)

            X_transform = self.zero_var_transformer.transform(X_transform)
            X_no_transform = self.zero_var_no_transformer.transform(X_no_transform)

            # Transform data if a transformer is supplied
            if isinstance(self.transformer, type(None)) == False:
                self.transformer = clone(self.transformer)

                X_transform = self.transformer.fit_transform(X_transform)

            # Recombine
            X_comb = np.hstack((X_transform, X_no_transform))

            return X_comb, y_comb

        # Prepare dataset if columns are excluded from transformation
        else:
            # Remove zero-variance features
            self.zero_var_transformer = VarianceThreshold().fit(X_comb)
            X_transform = self.zero_var_transformer.transform(X_comb)

            # Transform data if a transformer is supplied
            if isinstance(self.transformer, type(None)) == False:
                self.transformer = clone(self.transformer)

                # Transform data
                X_transform = self.transformer.fit_transform(X_transform)

            return X_transform, y_comb

    def transform(self, X):

        # Prepare dataset if columns are excluded from transformation
        if self.exclude_cols[0]:
            # Split data into columns excluded from transformation and included for transformation
            excl_range = np.asarray(self.exclude_cols[1])

            X_transform = np.delete(X, excl_range, axis=1)
            X_no_transform = X[:, excl_range]

            # Remove zero variance features
            X_transform = self.zero_var_transformer.transform(X_transform)
            X_no_transform = self.zero_var_no_transformer.transform(X_no_transform)

            # Transform data if a transformer is supplied
            if isinstance(self.transformer, type(None)) == False:
                X_transform = self.transformer.transform(X_transform)

            if X_transform.ndim == 1:
                X_transform = [X_transform]

            # Recombine
            return np.hstack((X_transform, X_no_transform))

        else:
            # Remove zero-variance features
            X_transform = self.zero_var_transformer.transform(X)

            # Transform data if a transformer is supplied
            if isinstance(self.transformer, type(None)) == False:
                X_transform = self.transformer.transform(X_transform)

            if X_transform.ndim == 1:
                X_transform = np.asarray([X_transform])

            return X_transform





from TreeOrdination import (
    TreeOrdination,
    CLRClosureTransformer,
    ResampleRandomizeTransform
)

from sklearn.datasets import make_classification
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.metrics import balanced_accuracy_score, median_absolute_error

from skbio.stats.composition import clr, closure, multiplicative_replacement

import pandas as pd

import numpy as np

from pathlib import Path

dirpath = Path(__file__).parent

# Tests the transformer modules
def test_transformers():

    # Create the dataset
    X, y = make_classification(
        n_samples=200,
        n_features=20,
        n_informative=5,
        n_redundant=2,
        n_repeated=0,
        n_classes=2,
        shuffle=False,
        random_state=0,
    )

    # To prevent negative proportions
    X_non_neg = np.abs(X)

    # Ensures that CLRTransformer returns the CLR Transform of X
    R = pd.DataFrame(CLRClosureTransformer(do_clr=True).fit_transform(X_non_neg))

    X_clr = pd.DataFrame(clr(multiplicative_replacement(closure(X_non_neg))))
    pd.testing.assert_frame_equal(X_clr, R, check_dtype=False)

    # Ensures that CLRTransformer returns the Closure of X
    R = pd.DataFrame(CLRClosureTransformer(do_clr=False).fit_transform(X_non_neg))
    X_closure = pd.DataFrame(closure(X_non_neg))
    pd.testing.assert_frame_equal(X_closure, R, check_dtype=False)

    # Ensure sampler, transformer objected works
    R_trf = ResampleRandomizeTransform(None, StandardScaler(), [False, -1])
    R, y_random = R_trf.fit_resample(X, y)
    
    assert R.shape[0] == int(X.shape[0] * 2)
    assert R.shape[1] == X.shape[1]
    assert y_random.shape[0] == int(X.shape[0] * 2)

# Tests the overall TreeOrdination pipeline
def test_treeord_basic():

    # Create the dataset
    X, y = make_classification(
        n_samples=200,
        n_features=20,
        n_informative=5,
        n_redundant=2,
        n_repeated=0,
        n_classes=2,
        shuffle=False,
        random_state=0,
    )

    # Split into train and test sets
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=0, stratify=y
    )

    # Set up TreeOrdinaton
    model = TreeOrdination(
        feature_names=[i for i in range(0, X.shape[1])],
        transformer=StandardScaler(),
        n_iter_unsup=2,
    )

    # Identify predictive features
    model.fit(X_train, y_train)

    X_emb_train = model.emb_transform(X_train, "PCA")
    X_approx_train = model.emb_transform(X_train, "approx")

    X_emb_test = model.emb_transform(X_test, "PCA")
    X_approx_test = model.emb_transform(X_test, "approx")

    error_train = median_absolute_error(X_emb_train, X_approx_train)
    assert error_train <= 0.75

    error_test = median_absolute_error(X_emb_test, X_approx_test)
    assert error_test <= 0.75

    p = model.predict(X_test)
    s = balanced_accuracy_score(y_test, p)
    assert s >= 0.75

    model.plot_projection(X, y)

    model.get_importance()
    model.plot_importance_global(X_test, y_test, 0, 10, 0)
    model.plot_importance_local(X_test[8], 10, 0)


test_transformers()
test_treeord_basic()



import numpy as np
 
from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
from sklearn.preprocessing import LabelEncoder
from sklearn.base import ClassifierMixin, BaseEstimator, clone
from sklearn.decomposition import PCA

from scipy.sparse import csr_array
from scipy.sparse import hstack as sp_hstack

from umap import UMAP

from LANDMark import LANDMarkClassifier

from matplotlib import pyplot as plt

import seaborn as sns

from .transformers_treeord import CLRClosureTransformer, ResampleRandomizeTransform
from .feature_importance_treeord import ExplainImportance


def basic_transform(X, transformer, exclude_col):

    if X.ndim == 2:
        X_in = np.copy(X, "C")

    else:
        X_in = np.asarray([X])

    # Transform data
    if exclude_col[0]:
        excl_range = np.asarray(exclude_col[1])

        X_transformed = np.delete(X_in, excl_range, axis=1)

        if isinstance(transformer, type(None)) == False:
            X_transformed = transformer.transform(X_transformed)

        X_no_transform = X_in[:, excl_range]

        if X_transformed.ndim == 1:
            X_transformed = X_transformed.reshape(-1,1)

        if X_no_transform.ndim == 1:
             X_no_transform = X_no_transform.rehsape(-1,1)

        X_transformed = np.hstack((X_transformed, X_no_transform))

    else:
        X_transformed = X_in

        if isinstance(transformer, type(None)) == False:
            X_transformed = transformer.transform(X_in)

    if X.ndim == 2:
        return X_transformed

    else:
        return X_transformed[0]


# Tree Ordination class
class TreeOrdination(ClassifierMixin, BaseEstimator):
    def __init__(
        self,
        feature_names,
        resampler=None,
        metric="hamming",
        prox_method = "terminal",
        supervised_clf=ExtraTreesClassifier(1024),
        proxy_model = ExtraTreesRegressor(1024),
        landmark_model = LANDMarkClassifier(160, use_nnet=False, n_jobs = 8),
        n_iter_unsup=5,
        transformer=None,
        exclude_col=[False, 0],
        n_neighbors=8,
        n_components=2,
        min_dist=0.001,
    ):
        self.feature_names = feature_names

        self.resampler = resampler

        self.metric = metric
        self.prox_method = prox_method

        self.supervised_clf = supervised_clf
        self.proxy_model = proxy_model
        self.landmark_model = landmark_model

        self.n_iter_unsup = n_iter_unsup

        self.transformer = transformer

        self.exclude_col = exclude_col

        self.n_neighbors = n_neighbors
        self.n_components = n_components
        self.min_dist = min_dist

    def get_initial_embedding(self, X):

        # Get an Initial LANDMark Representations
        self.Rs = []
        self.LM_emb = []
        self.transformers = []

        for i in range(self.n_iter_unsup):

            print("Iteration %d..." %i)

            #No resample and transform
            resampler = ResampleRandomizeTransform(self.resampler,
                                                   self.transformer,
                                                   self.exclude_col)

            X_rand, y_rand = resampler.fit_resample(X, self.y)

            # Train model
            model = clone(self.landmark_model)

            model.fit(X_rand, y_rand)
            self.Rs.append(model)

            # Get proximity
            X_trf = resampler.transform(X)

            # Update Overall Proximity
            if i > 0:
                self.LM_emb = sp_hstack((self.LM_emb, model.proximity(X_trf, self.prox_method)))
            else:
                self.LM_emb = model.proximity(X_trf, self.prox_method)

            # Save the resampler
            self.transformers.append(resampler)

        # Get Embeddings
        self.UMAP_trf = UMAP(
            n_neighbors=self.n_neighbors,
            n_components=15,
            min_dist=self.min_dist,
            metric="hamming",
            densmap=False,
        ).fit(self.LM_emb)

        self.UMAP_emb = self.UMAP_trf.transform(self.LM_emb)

        self.PCA_trf = PCA(self.n_components, whiten = True).fit(self.UMAP_emb)

        self.PCA_emb = self.PCA_trf.transform(self.UMAP_emb)

    def fit(self, X, y=None):

        # Save X and y
        self.X = np.copy(X, "C")
        self.y = np.copy(y, "C")

        # Encode y
        self.encoded_labels = LabelEncoder().fit(self.y)
        y_enc = self.encoded_labels.transform(self.y)

        # Get initial embeddings, features, and cluster
        self.f_name = np.asarray(self.feature_names)

        self.get_initial_embedding(X)

        # Train a Classification model
        self.p_model = clone(self.supervised_clf).fit(self.LM_emb, y_enc.astype(int))

        # Train a projection model (Approximate Transformer)
        if self.exclude_col[0]:
            excl_range = np.asarray(self.exclude_col[1])

            X_transformed = np.delete(X, excl_range, axis=1)

            if isinstance(self.transformer, type(None)) == False:
                self.proj_transformer = clone(self.transformer)

                X_transformed = self.proj_transformer.fit_transform(X_transformed)

            else:
                self.proj_transformer = None

            X_no_transform = X[:, excl_range]

            X_transformed = np.hstack((X_transformed, X_no_transform))

        else:
            X_transformed = np.copy(X, "C")

            if isinstance(self.transformer, type(None)) == False:
                self.proj_transformer = clone(self.transformer)

                X_transformed = self.proj_transformer.fit_transform(X)

            else:
                self.proj_transformer = None

        self.l_model = clone(self.proxy_model).fit(
            X_transformed, self.PCA_emb
        )

        return self

    def get_importance(self):

        self.feature_importance_explainer = ExplainImportance(self.l_model, self.f_name)

        self.feature_importance_explainer.get_importance()

    def plot_importance_global(
        self, X, y, class_name, n_features=10, axis=0, summary="median", **kwargs
    ):

        # Transform Samples
        X_transformed = basic_transform(X, self.proj_transformer, self.exclude_col)

        # Grab samples for a specific class, if supplied
        y_locs = np.where(y == class_name, True, False)

        return self.feature_importance_explainer.plot_importance(
            X_transformed[y_locs], n_features, class_name, axis, summary, **kwargs
        )

    def plot_importance_local(self, X, n_features=10, axis=0, **kwargs):

        # Transform data
        X_transformed = basic_transform(X, self.proj_transformer, self.exclude_col)

        return self.feature_importance_explainer.plot_importance(
            X_transformed,
            n_features,
            class_name=None,
            axis=axis,
            summary=None,
            **kwargs
        )

    def plot_projection(self, X, y, ax_1=0, ax_2=1, use_approx=True, trf_type = "PCA"):

        # Plot data
        projection = self.emb_transform(X, trf_type)

        fig, ax = plt.subplots(figsize=(10, 5))

        sns.scatterplot(x=projection[:, ax_1], y=projection[:, ax_2], hue=y, ax=ax)

        pc_ax_1 = self.PCA_trf.explained_variance_ratio_[ax_1] * 100
        pc_ax_2 = self.PCA_trf.explained_variance_ratio_[ax_2] * 100

        ax.set_xlabel("PCA %s (%.3f Percent)" % (str(ax_1 + 1), pc_ax_1))
        ax.set_ylabel("PCA %s (%.3f Percent)" % (str(ax_2 + 1), pc_ax_2))

        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))

        return fig, ax

    def predict_proba(self, X):

        tree_emb = self.emb_transform(X, "UMAP")

        P = self.p_model.predict_proba(tree_emb)

        return P

    def predict(self, X):

        tree_emb = self.emb_transform(X, "LM")

        P = self.p_model.predict(tree_emb)

        P = self.encoded_labels.inverse_transform(P)

        return P

    def emb_transform(self, X, trf_type = "PCA"):

        if trf_type == "approx":
            X_transformed = basic_transform(X, self.proj_transformer, self.exclude_col)

            return self.l_model.predict(X_transformed)

        else:
            tree_emb = []

            for i in range(self.n_iter_unsup):

                # Get trained model
                model = self.Rs[i]
                transformer = self.transformers[i]

                # Get proximity
                if i != 0:
                    tree_emb = sp_hstack((tree_emb, model.proximity(transformer.transform(X), self.prox_method)))
                else:
                    tree_emb = model.proximity(transformer.transform(X), self.prox_method)

            if trf_type == "LM":
                return tree_emb

            umap_emb = self.UMAP_trf.transform(tree_emb)
            if trf_type == "UMAP":
                return umap_emb

            pca_emb = self.PCA_trf.transform(umap_emb)
            if trf_type == "PCA":
                return pca_emb



from shap import Explainer

import numpy as np

import matplotlib.pyplot as plt

import pandas as pd


class ExplainImportance:
    """
    Returns an Explanation Model.
    """

    def __init__(self, model, feature_names):

        self.model = model
        self.feature_names = feature_names

    def get_importance(self):

        # TO-DO: Add KernelExplainer for models not based on trees.
        #        Add explainer for neural networks

        self.explainer = Explainer(self.model)

    def plot_importance(
        self, X, n_features, class_name, axis=0, summary="median", **kwargs
    ):
        """
        Create a horizontal barchart of feature effects, sorted by their magnitude.

        Adapted from: https://docs.seldon.io/projects/alibi/en/stable/examples/path_dependent_tree_shap_adult_xgb.html
        """
        # Calculate importances
        feat_imp = self.explainer.shap_values(X)[axis]

        # Determine if Global or Local feature importances are to be printed
        # Determine how SHAP values are to be sumarized across multiple samples from each class
        if feat_imp.ndim == 2:
            overview = "Global"

            ylabel = "Feature Name"

            if summary == "median":
                feat_imp = np.median(feat_imp, axis=0)
                xlabel = "Median(SHAP Values)"

            elif summary == "mean":
                feat_imp = np.mean(feat_imp, axis=0)
                xlabel = "Mean(SHAP Values)"

            elif summary == "abs_median":
                feat_imp = np.median(np.abs(feat_imp), axis=0)
                xlabel = "Median(|SHAP Values|)"

            elif summary == "abs_mean":
                feat_imp = np.mean(np.abs(feat_imp), axis=0)
                xlabel = "Mean(|SHAP Values|)"

        elif feat_imp.ndim == 1:
            overview = "Local"
            xlabel = "SHAP Values"
            ylabel = "Feature Name (Magnitude)"
            prediction = self.model.predict([X])[0][axis]

        # Sort Features
        feat_imp_pd = pd.Series(data=feat_imp, index=self.feature_names)
        feat_imp_sorted_index = (
            feat_imp_pd.abs().sort_values(ascending=False).index.values[0:n_features]
        )
        feat_imp_sorted_data = feat_imp_pd[feat_imp_sorted_index].values

        if X.ndim == 1:
            feat_values = pd.Series(data=X, index=self.feature_names)[
                feat_imp_sorted_index
            ].values

            feat_imp_sorted_index = np.asarray(
                [
                    "%s (%.2f)" % (f_name, feat_values[i])
                    for i, f_name in enumerate(feat_imp_sorted_index)
                ]
            )

        # Get colors of bars
        bar_cols = [
            "Orange" if score < 0 else "Green" for score in feat_imp_sorted_data
        ]

        # Some plot aesthetics
        labels_fontsize = kwargs.get("labels_fontsize", 10)
        tick_labels_fontsize = kwargs.get("tick_labels_fontsize", 10)

        # plot
        fig, ax = plt.subplots(figsize=(10, 5))
        y_pos = np.arange(len(feat_imp_sorted_data))
        ax.barh(y_pos, feat_imp_sorted_data, color=bar_cols)

        # set lables
        ax.set_yticks(y_pos)
        ax.set_yticklabels(feat_imp_sorted_index, fontsize=tick_labels_fontsize)
        ax.invert_yaxis()
        ax.set_xlabel(xlabel, fontsize=labels_fontsize)
        ax.set_ylabel(ylabel, fontsize=labels_fontsize)

        if overview == "Global":
            plt.title(
                "TreeOrdination %s Feature Importance Plot for Class %s"
                % (overview, str(class_name))
            )

        else:
            plt.title(
                "TreeOrdination %s Feature Importance Plot\nPredicted Model Output for Sample = %.2f"
                % (overview, prediction)
            )

        return ax, fig



"""
This module imports TreeOrdination
"""

from .TreeOrdination import TreeOrdination
from .transformers_treeord import CLRClosureTransformer, ResampleRandomizeTransform

__all__ = ["TreeOrdination", "CLRClosureTransformer", "ResampleRandomizeTransform"]
